==> gmssl-jsse-provider-0.1.4-20190730.152528-2-sources.jar: org/bouncycastle/tls/DeferredHash.java <==
package org.bouncycastle.tls;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Enumeration;
import java.util.Hashtable;

import org.bouncycastle.tls.crypto.TlsHash;
import org.bouncycastle.tls.crypto.impl.bc.BcTlsCryptoSdf;
import org.bouncycastle.util.Shorts;

/**
 * Buffers input until the hash algorithm is determined.
 */
class DeferredHash
        implements TlsHandshakeHash {
    protected static final int BUFFERING_HASH_LIMIT = 4;

    protected TlsContext context;

    private DigestInputBuffer buf;
    private Hashtable<Short, TlsHash> hashes;
    private boolean forceBuffering;
    private boolean sealed;

    DeferredHash(TlsContext context) {
        this.context = context;
        this.buf = new DigestInputBuffer();
        this.hashes = new Hashtable();
        this.forceBuffering = false;
        this.sealed = false;
    }

    private DeferredHash(TlsContext context, Hashtable hashes) {
        this.context = context;
        this.buf = null;
        this.hashes = hashes;
        this.forceBuffering = false;
        this.sealed = true;
    }

    @Override
    public void copyBufferTo(OutputStream output) throws IOException {
        if (buf == null) {
            // If you see this, you need to call forceBuffering() before sealHashAlgorithms()
            throw new IllegalStateException("Not buffering");
        }

        buf.copyTo(output);
    }

    @Override
    public void forceBuffering() {
        if (sealed) {
            throw new IllegalStateException("Too late to force buffering");
        }

        this.forceBuffering = true;
    }

    @Override
    public TlsHandshakeHash notifyPRFDetermined() {
        int prfAlgorithm = context.getSecurityParameters().getPrfAlgorithm();
        if (prfAlgorithm == PRFAlgorithm.tls_prf_legacy) {
            checkTrackingHash(HashAlgorithm.md5);
            checkTrackingHash(HashAlgorithm.sha1);
        } else {
            checkTrackingHash(TlsUtils.getHashAlgorithmForPRFAlgorithm(prfAlgorithm));
        }
        return this;
    }

    @Override
    public void trackHashAlgorithm(short hashAlgorithm) {
        if (sealed) {
            throw new IllegalStateException("Too late to track more hash algorithms");
        }

        checkTrackingHash(Shorts.valueOf(hashAlgorithm));
    }

    @Override
    public void sealHashAlgorithms() {
        if (!sealed) {
            sealed = true;
            checkStopBuffering();
        }
    }

    @Override
    public TlsHandshakeHash stopTracking() {
        Hashtable newHashes = new Hashtable();
        int prfAlgorithm = context.getSecurityParameters().getPrfAlgorithm();
        if (prfAlgorithm == PRFAlgorithm.tls_prf_legacy) {
            cloneHash(newHashes, HashAlgorithm.md5);
            cloneHash(newHashes, HashAlgorithm.sha1);
        } else {
            cloneHash(newHashes, TlsUtils.getHashAlgorithmForPRFAlgorithm(prfAlgorithm));
        }
        this.reset();
        return new DeferredHash(context, newHashes);
    }

    @Override
    public TlsHash forkPRFHash() {
        checkStopBuffering();

        TlsHash prfHash;

        int prfAlgorithm = context.getSecurityParameters().getPrfAlgorithm();
        if (prfAlgorithm == PRFAlgorithm.tls_prf_legacy) {
            prfHash = new CombinedHash(context, cloneHash(HashAlgorithm.md5), cloneHash(HashAlgorithm.sha1));
        } else {
            prfHash = cloneHash(TlsUtils.getHashAlgorithmForPRFAlgorithm(prfAlgorithm));
        }

        if (buf != null) {
            buf.updateDigest(prfHash);
        }

        return prfHash;
    }

    @Override
    public byte[] getFinalHash(short hashAlgorithm) {
        TlsHash d = (TlsHash) hashes.get(Shorts.valueOf(hashAlgorithm));
        if (d == null) {
            throw new IllegalStateException("HashAlgorithm." + HashAlgorithm.getText(hashAlgorithm) + " is not being tracked");
        }

        d = (TlsHash) d.clone();
        if (buf != null) {
            buf.updateDigest(d);
        }

        return d.calculateHash();
    }

    @Override
    public void update(byte[] input, int inOff, int len) {
        if (buf != null) {
            buf.write(input, inOff, len);
            return;
        }

        Enumeration e = hashes.elements();
        while (e.hasMoreElements()) {
            TlsHash hash = (TlsHash) e.nextElement();
            hash.update(input, inOff, len);
        }
    }

    @Override
    public byte[] calculateHash() {
        throw new IllegalStateException("Use fork() to get a definite Digest");
    }

    @Override
    public Object clone() {
        throw new IllegalStateException("attempt to clone a DeferredHash");
    }

    @Override
    public void reset() {
        this.forceBuffering = false;
        this.sealed = false;

        if (buf != null) {
            buf.reset();
            return;
        }

        Enumeration e = hashes.elements();
        while (e.hasMoreElements()) {
            TlsHash hash = (TlsHash) e.nextElement();
            hash.reset();
        }
    }

    public void releaseConnection(){
        Enumeration e = hashes.elements();
        while (e.hasMoreElements()) {
            TlsHash hash = (TlsHash) e.nextElement();
            if (hash instanceof BcTlsCryptoSdf.BcTlsHash){
                ((BcTlsCryptoSdf.BcTlsHash) hash).releaseConnection();
            }
        }
    }

    protected void checkStopBuffering() {
        if (!forceBuffering && sealed && buf != null && hashes.size() <= BUFFERING_HASH_LIMIT) {
            Enumeration e = hashes.elements();
            while (e.hasMoreElements()) {
                TlsHash hash = (TlsHash) e.nextElement();
                buf.updateDigest(hash);
            }

            this.buf = null;
        }
    }

    protected void checkTrackingHash(Short hashAlgorithm) {
        if (!hashes.containsKey(hashAlgorithm)) {
            TlsHash hash = context.getCrypto().createHash(hashAlgorithm.shortValue());
            hashes.put(hashAlgorithm, hash);
        }
    }

    protected TlsHash cloneHash(Short hashAlgorithm) {
        TlsHash tlsHash = hashes.get(hashAlgorithm);
        return (TlsHash) tlsHash.clone();
    }

    protected void cloneHash(Hashtable newHashes, Short hashAlgorithm) {
        TlsHash hash = cloneHash(hashAlgorithm);
        if (buf != null) {
            buf.updateDigest(hash);
        }
        newHashes.put(hashAlgorithm, hash);
    }
}

==> gmssl-jsse-provider-1.0.4-20200525.114049-10-sources.jar: org/bouncycastle/tls/DeferredHash.java <==
package org.bouncycastle.tls;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Enumeration;
import java.util.Hashtable;

import org.bouncycastle.tls.crypto.TlsHash;
import org.bouncycastle.tls.crypto.impl.bc.BcTlsCryptoSdf;
import org.bouncycastle.util.Shorts;

/**
 * Buffers input until the hash algorithm is determined.
 */
class DeferredHash implements TlsHandshakeHash {
    protected static final int BUFFERING_HASH_LIMIT = 4;

    protected TlsContext context;

    private DigestInputBuffer buf;
    private Hashtable<Short, TlsHash> hashes;
    private boolean forceBuffering;
    private boolean sealed;

    DeferredHash(TlsContext context) {
        this.context = context;
        this.buf = new DigestInputBuffer();
        this.hashes = new Hashtable();
        this.forceBuffering = false;
        this.sealed = false;
    }

    private DeferredHash(TlsContext context, Hashtable hashes) {
        this.context = context;
        this.buf = null;
        this.hashes = hashes;
        this.forceBuffering = false;
        this.sealed = true;
    }

    @Override
    public void copyBufferTo(OutputStream output) throws IOException {
        if (buf == null) {
            // If you see this, you need to call forceBuffering() before sealHashAlgorithms()
            throw new IllegalStateException("Not buffering");
        }

        buf.copyTo(output);
    }

    @Override
    public void forceBuffering() {
        if (sealed) {
            throw new IllegalStateException("Too late to force buffering");
        }

        this.forceBuffering = true;
    }

    @Override
    public TlsHandshakeHash notifyPRFDetermined() {
        int prfAlgorithm = context.getSecurityParameters().getPrfAlgorithm();
        if (prfAlgorithm == PRFAlgorithm.tls_prf_legacy) {
            checkTrackingHash(HashAlgorithm.md5);
            checkTrackingHash(HashAlgorithm.sha1);
        } else {
            checkTrackingHash(TlsUtils.getHashAlgorithmForPRFAlgorithm(prfAlgorithm));
        }
        return this;
    }

    @Override
    public void trackHashAlgorithm(short hashAlgorithm) {
        if (sealed) {
            throw new IllegalStateException("Too late to track more hash algorithms");
        }

        checkTrackingHash(Shorts.valueOf(hashAlgorithm));
    }

    @Override
    public void sealHashAlgorithms() {
        if (!sealed) {
            sealed = true;
            checkStopBuffering();
        }
    }

    @Override
    public TlsHandshakeHash stopTracking() {
        Hashtable newHashes = new Hashtable();
        int prfAlgorithm = context.getSecurityParameters().getPrfAlgorithm();
        if (prfAlgorithm == PRFAlgorithm.tls_prf_legacy) {
            cloneHash(newHashes, HashAlgorithm.md5);
            cloneHash(newHashes, HashAlgorithm.sha1);
        } else {
            cloneHash(newHashes, TlsUtils.getHashAlgorithmForPRFAlgorithm(prfAlgorithm));
        }
        this.releaseConnection();
        return new DeferredHash(context, newHashes);
    }

    @Override
    public TlsHash forkPRFHash() {
        checkStopBuffering();

        TlsHash prfHash;

        int prfAlgorithm = context.getSecurityParameters().getPrfAlgorithm();
        if (prfAlgorithm == PRFAlgorithm.tls_prf_legacy) {
            prfHash = new CombinedHash(context, cloneHash(HashAlgorithm.md5), cloneHash(HashAlgorithm.sha1));
        } else {
            prfHash = cloneHash(TlsUtils.getHashAlgorithmForPRFAlgorithm(prfAlgorithm));
        }

        if (buf != null) {
            buf.updateDigest(prfHash);
        }

        return prfHash;
    }

    @Override
    public byte[] getFinalHash(short hashAlgorithm) {
        TlsHash d = (TlsHash) hashes.get(Shorts.valueOf(hashAlgorithm));
        if (d == null) {
            throw new IllegalStateException("HashAlgorithm." + HashAlgorithm.getText(hashAlgorithm) + " is not being tracked");
        }

        d = (TlsHash) d.clone();
        if (buf != null) {
            buf.updateDigest(d);
        }

        return d.calculateHash();
    }

    @Override
    public void update(byte[] input, int inOff, int len) {
        if (buf != null) {
            buf.write(input, inOff, len);
            return;
        }

        Enumeration e = hashes.elements();
        while (e.hasMoreElements()) {
            TlsHash hash = (TlsHash) e.nextElement();
            hash.update(input, inOff, len);
        }
    }

    @Override
    public byte[] calculateHash() {
        throw new IllegalStateException("Use fork() to get a definite Digest");
    }

    @Override
    public Object clone() {
        throw new IllegalStateException("attempt to clone a DeferredHash");
    }

    @Override
    public void reset() {
        this.forceBuffering = false;
        this.sealed = false;

        if (buf != null) {
            buf.reset();
            return;
        }

        Enumeration e = hashes.elements();
        while (e.hasMoreElements()) {
            TlsHash hash = (TlsHash) e.nextElement();
            hash.reset();
        }
    }

    public void releaseConnection() {
        Enumeration e = hashes.elements();
        while (e.hasMoreElements()) {
            TlsHash hash = (TlsHash) e.nextElement();
            if (hash instanceof BcTlsCryptoSdf.BcTlsHash) {
                ((BcTlsCryptoSdf.BcTlsHash) hash).releaseConnection();
            }
        }
    }

    protected void checkStopBuffering() {
        if (!forceBuffering && sealed && buf != null && hashes.size() <= BUFFERING_HASH_LIMIT) {
            Enumeration e = hashes.elements();
            while (e.hasMoreElements()) {
                TlsHash hash = (TlsHash) e.nextElement();
                buf.updateDigest(hash);
            }

            this.buf = null;
        }
    }

    protected void checkTrackingHash(Short hashAlgorithm) {
        if (!hashes.containsKey(hashAlgorithm)) {
            TlsHash hash = context.getCrypto().createHash(hashAlgorithm.shortValue());
            hashes.put(hashAlgorithm, hash);
        }
    }

    protected TlsHash cloneHash(Short hashAlgorithm) {
        TlsHash tlsHash = hashes.get(hashAlgorithm);
        return (TlsHash) tlsHash.clone();
    }

    protected void cloneHash(Hashtable newHashes, Short hashAlgorithm) {
        TlsHash hash = cloneHash(hashAlgorithm);
        if (buf != null) {
            buf.updateDigest(hash);
        }
        newHashes.put(hashAlgorithm, hash);
    }
}
